query GetPr($repository: String!, $owner: String!, $labels: [String!], $after: String) {
    repository(name: $repository, owner: $owner) {
        pullRequests(
            states: MERGED,
            labels: $labels,
            first: 100,
            orderBy: {
                field: UPDATED_AT
                direction: DESC
            }
            baseRefName: "main",
            after: $after
        ) {
            totalCount,
            pageInfo {
                hasNextPage,
                endCursor
            }
            edges {
                node {
                    labels(first: 10) {
                        nodes {
                            name
                        }
                    },
                    author {
                        login
                    }
                    number,
                    title,
                    closedAt
                }
            }
        }
    }
}
